#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <glib.h>
#include <iconv.h>
#include <locale.h>
#include <assert.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>


/*
#include "../config.h"
#ifdef HAVE_CODESET
#include <langinfo.h>
#endif
*/

#include "charset.h"
#include "charset_config.h"

#ifdef AUTO_ENCA_SUPPORT
#ifdef DYNAMIC_ENCA
#include <charset_fake_enca.h>
#else
#include <enca.h>
#endif
#endif

#define LOCALE_STRING_LENGTH 64
static int xmms_force_new = 1;

static char charset_default[LOCALE_STRING_LENGTH];
static int current_language = -2;
static int default_language;
static char xmms_charset_tmp[CHARSET_MAX_STRING_SIZE];

static int xmms_charset_changed = -1;

static int xmms_charset_id3;
static int xmms_charset_output;
static int xmms_charset_fs;
static int xmms_charset_pl;

static iconv_t iconv_id3=(iconv_t)-1;	/* id3 to output */
static iconv_t iconv_output=(iconv_t)-1; /* output to id3 */
static iconv_t iconv_fs=(iconv_t)-1; /* fs to id3 */
static iconv_t iconv_fsout=(iconv_t)-1; /* fs to out */
static iconv_t iconv_utf=(iconv_t)-1; /* id3 to UTF-8 */
static iconv_t iconv_pl2fs=(iconv_t)-1; /* playlist to fs */
static iconv_t iconv_fs2pl=(iconv_t)-1; /* fs to playlist */
static iconv_t iconv_ct2out=(iconv_t)-1; /* LC_CTYPE to output */

static int xmms_autocharset_mode = 0;
static int xmms_autocharset_fs = 1;
static int xmms_charset_fm = 1;

static int xmms_autocharset_changed = -1;
static iconv_t *xmms_autocharset = NULL;
static iconv_t *xmms_autocharset_id2utf = NULL;
static iconv_t *xmms_autocharset_id2id = NULL;
static iconv_t *xmms_autocharset_id2fs = NULL;
static iconv_t *xmms_autocharset_fs2id = NULL;
static int xmms_autocharset_number = -1;

//static int recode_fsout;
static int recode_fs2utf;
#include "charset_auto_russian.h"

extern int errno;

static int sure_default;

/*
static PluginOptions *po = NULL;
*/

/*****************************************************************************
************************ Support Functions ***********************************
*****************************************************************************/

const char *charset_get_output_charset() {
    if (xmms_charset_output) return charset_list[xmms_charset_output];
    else return charset_default;
}


const charset_t *get_charset_list(int csl) {
    if (!csl) return charset_list_default;
    if (csl==1) {
	if (default_language>=0) return languages[default_language].enc;
	else return charset_list_default;
    }
    return languages[csl-2].enc;
}

const autocharset_t *get_autocharset_list(int csl, int *ni) {
    if (!csl) {
	*ni=autocharset_list_ni_default;
	return autocharset_list_default;
    }
    if (csl==1) {
	if (default_language>=0) {
	    *ni=languages[default_language].n_autoenc;
	    return languages[default_language].autoenc;
	}
	else {
	    *ni=autocharset_list_ni_default;
	    return autocharset_list_default;
	}
    }
    *ni=languages[csl-2].n_autoenc;
    return languages[csl-2].autoenc;
}


int get_default_language() {
    int i;
    char *l, ltmp[3];
    l=setlocale(ENV_VARIABLE,"");
    if (!l) return -1;
    else if ((strcmp(l,"C")==0)||(strcmp(l,"POSIX")==0)) return -1;
    else {
	ltmp[2]=0;
	for (i=0;language_aliases[i].alias;i++) 
	    if (strncmp(l,language_aliases[i].alias,5)==0) {
		strncpy(ltmp,language_aliases[i].lang,2);
		break;
	    }
	if (!language_aliases[i].alias) strncpy(ltmp,l,2);
	for (i=0;languages[i].sn;i++)
	    if (strcmp(ltmp,languages[i].sn)==0) return i;
    }
    return -1;
}

static int xmms_charset_number(char *c) {
    int i;
    if (!c) return 0;
    for (i=0;charset_list[i];i++)
	if (strcmp(charset_list[i], c)==0) return i;
    return 0; /* Default, - if not found */
}

/*****************************************************************************
************************** Initialization  ***********************************
*****************************************************************************/

#ifdef AUTO_ENCA_SUPPORT
static EncaAnalyser enca = NULL;
static int enca_pos=-1;

#ifdef DYNAMIC_ENCA
# include <dlfcn.h>
#ifndef RTLD_NOW
# define RTLD_NOW 0
#endif
static void *enca_handle = NULL;
static int dynamic_enca_initialised = 0;

static void* open_dynamic_lib(char *filename)
{
    return dlopen(filename, RTLD_NOW);
}

static void close_dynamic_lib(void* handle)
{
    dlclose(handle);
}

static void* find_dynamic_symbol(void *handle, char *symbol)
{
#ifdef SYMBOL_PREFIX
	char *sym = g_strconcat(SYMBOL_PREFIX, symbol, NULL);
	void *symh = dlsym(handle, sym);
	g_free(sym);
	return symh;
#else
	return dlsym(handle, symbol);
#endif
}
#endif
/*
    enca_analyse_const(enca,buf,len);
    enca_charset_name(ee.charset, ENCA_NAME_STYLE_ICONV);
    enca_analyser_alloc(ltmp);
    enca_analyser_free(enca);
    enca_set_threshold(enca, 1);
    enca_set_multibyte(enca, 1);
    enca_set_ambiguity(enca, 1);
    enca_set_garbage_test(enca, 0);
    enca_set_filtering(enca, 0);
    enca_set_significant(enca,1);
    enca_set_termination_strictness(enca,0);
    enca_get_language_charsets (const char *langname,
                                          size_t *n);
*/

#ifdef DYNAMIC_ENCA
static void charset_dynamic_enca_init() {
    if (dynamic_enca_initialised) return;
#ifdef _CHARSET_DEBUG
    fprintf(stderr,"charset_dynamic_enca_init\n");
#endif
    dynamic_enca_initialised=1;
    enca_handle=open_dynamic_lib(ENCA_LIB);
    if (enca_handle) {
	enca_set_multibyte=find_dynamic_symbol(enca_handle,"enca_set_multibyte");
	enca_set_interpreted_surfaces=find_dynamic_symbol(enca_handle,"enca_set_interpreted_surfaces");
	enca_set_ambiguity=find_dynamic_symbol(enca_handle,"enca_set_ambiguity");
	enca_set_filtering=find_dynamic_symbol(enca_handle,"enca_set_filtering");
	enca_set_garbage_test=find_dynamic_symbol(enca_handle,"enca_set_garbage_test");
	enca_set_termination_strictness=find_dynamic_symbol(enca_handle,"enca_set_termination_strictness");
	enca_set_significant=find_dynamic_symbol(enca_handle,"enca_set_significant");
	enca_set_threshold=find_dynamic_symbol(enca_handle,"enca_set_threshold");
	enca_charset_name=find_dynamic_symbol(enca_handle,"enca_charset_name");
	enca_get_language_charsets=find_dynamic_symbol(enca_handle,"enca_get_language_charsets");
	enca_analyser_alloc=find_dynamic_symbol(enca_handle,"enca_analyser_alloc");
	enca_analyser_free=find_dynamic_symbol(enca_handle,"enca_analyser_free");
	enca_analyse_const=find_dynamic_symbol(enca_handle,"enca_analyse_const");
	if ((!enca_set_multibyte)||(!enca_set_interpreted_surfaces)||(!enca_set_ambiguity)||
	   (!enca_set_filtering)||(!enca_set_garbage_test)||(!enca_set_termination_strictness)||
	   (!enca_set_significant)||(!enca_set_threshold)||(!enca_charset_name)||
	   (!enca_get_language_charsets)||(!enca_analyser_alloc)||(!enca_analyser_free)||
	   (!enca_analyse_const)) {
	    close_dynamic_lib(enca_handle);
	    enca_handle=NULL;
	    perror( "Enca dynamic: Incomplete function set" );
	}
    }
}
#endif

static void autocharset_enca_free() {
    int i,j;

#ifdef DYNAMIC_ENCA
    if (!enca_handle) return;
#endif
    if (!enca) return;
#ifdef _CHARSET_DEBUG
    fprintf(stderr,"autocharset_enca_free\n");
#endif

    for (i=0;i<autocharset_list_ni;i++)
        if (!strcmp(autocharset_list[i].title,"Enca Library")) {
	    g_free(autocharset_list[i].title);
	    for (j=0;autocharset_list[i].encodings[j];j++) g_free(autocharset_list[i].encodings[j]);
	    autocharset_list[i].encodings[0]=NULL;
	    autocharset_list_ni--;
	    break;
	}

    enca_analyser_free(enca);
    enca=NULL;
}

static void autocharset_enca_close() {
#ifdef DYNAMIC_ENCA
    if (enca_handle) {
#endif
#ifdef _CHARSET_DEBUG
	fprintf(stderr,"autocharset_enca_close\n");
#endif
	autocharset_enca_free();
#ifdef DYNAMIC_ENCA
	close_dynamic_lib(enca_handle);
	enca_handle=NULL;
	dynamic_enca_initialised=0;
    }
#endif
}


static void autocharset_enca_init(int cl) {
    int i;
    int n_charsets;
    int *charsets;

    if (cl<0) return;
#ifdef DYNAMIC_ENCA
    if (!enca_handle) return;
#endif
    if (enca) return;
    

#ifdef _CHARSET_DEBUG
    fprintf(stderr,"autocharset_enca_init\n");
#endif

	if (autocharset_list_ni<CHARSET_MAX_AUTOENCODINGS) {
	    for (i=0;i<autocharset_list_ni;i++)
		if (!strcmp(autocharset_list[i].title,"Enca Library")) break;
	    if (i==autocharset_list_ni) {
	    	enca=enca_analyser_alloc(languages[cl].sn);
		if (enca) {
		    enca_set_threshold(enca, 1);
		    enca_set_multibyte(enca, 1);
		    enca_set_ambiguity(enca, 1);
		    enca_set_garbage_test(enca, 0);
		    enca_set_filtering(enca, 0);
		    enca_set_significant(enca,1);
		    enca_set_termination_strictness(enca,0);

		    autocharset_list[autocharset_list_ni].title=g_strdup("Enca Library");
		    if (autocharset_list[autocharset_list_ni].title) {
			charsets=enca_get_language_charsets(languages[cl].sn, &n_charsets);
			if (charsets) {
			    if (n_charsets+2>CHARSET_MAX_ENCODINGS) n_charsets=CHARSET_MAX_ENCODINGS;
			    autocharset_list[autocharset_list_ni].encodings[0]="UTF-8";
			    for (i=1;i<n_charsets+1;i++) {
				autocharset_list[autocharset_list_ni].encodings[i]=g_strdup(enca_charset_name(charsets[i-1], ENCA_NAME_STYLE_ICONV));
			    }
			    free(charsets);
			    autocharset_list[autocharset_list_ni].encodings[n_charsets+1]=0;
			    autocharset_list[autocharset_list_ni].func=&autocharset_enca;
			    enca_pos=autocharset_list_ni++;
#ifdef _CHARSET_DEBUG
			    fprintf(stderr,"enca engine added\n");
#endif
			} else autocharset_list[autocharset_list_ni].encodings[0]=0;
		    } else {
			enca_analyser_free(enca);
			enca=NULL;
		    }
		}
	    }
	}
}

#endif

static void xmms_charset_setdefault_charset() {
	// Detecting default charset    
/*
My routine is better!
#ifdef HAVE_CODESET
    char *cs;
    cs=nl_langinfo(CODESET);
    if (cs) {
	strncpy(charset_default,cs,LOCALE_STRING_LENGTH);
	charset_default[LOCALE_STRING_LENGTH-1]=0;
    } else {
	strcpy(charset_default,"UTF-8");
    }
#else
*/
    charset_list_t *enc;
    char *str1, *str2;
    
    str1 = setlocale(ENV_VARIABLE,"");
    if (str1) {
        str2 = strrchr(str1,'.');
        if (str2) strncpy(charset_default,str2+1,LOCALE_STRING_LENGTH);
	else {
	    if (default_language<0) {
		strcpy(charset_default,"UTF-8");
		sure_default=0;
		return;
	    } else {
		enc=&(languages[default_language].enc);
		if ((enc[0])&&(enc[1]))
		    strncpy(charset_default,languages[default_language].enc[1],LOCALE_STRING_LENGTH);
		else {
		    strcpy(charset_default,"UTF-8");
		    sure_default=0;
		}
	    }
	}
	charset_default[LOCALE_STRING_LENGTH-1]=0;
        str2 = strchr(charset_default,'@');
        if (str2) *str2 = 0;
	sure_default=1;
    } else {
        strcpy(charset_default,"UTF-8");
	sure_default=0;
    }
#ifdef _CHARSET_DEBUG
    fprintf(stderr,"xmms_charset_setdefault_charset: %s\n",charset_default);
#endif
/* #endif */
}


static void xmms_autocharset_iconv_close();
static void xmms_charset_iconv_close();

static void xmms_autocharset_init(int m, int fm, int fs) {
    if (m<autocharset_list_ni) xmms_autocharset_mode=m;
    else if (autocharset_list_ni>1) xmms_autocharset_mode=1;
    else xmms_autocharset_mode=0;
    xmms_charset_fm=fm;
    xmms_autocharset_fs=fs;
#ifdef _CHARSET_DEBUG
    fprintf(stderr,"xmms_autocharset_init: %i\n",xmms_autocharset_mode);
#endif
}

static void xmms_autocharset_free() {
#ifdef _CHARSET_DEBUG
    fprintf(stderr,"xmms_autocharset_free\n");
#endif
    xmms_autocharset_iconv_close();
#ifdef AUTO_ENCA_SUPPORT
    autocharset_enca_close();
#endif
}


void xmms_charset_init(char *lng, char *id3, char *output, char *fs, char *pl, int am, int afm, int afs) {
    int i=-1;
#ifdef _CHARSET_DEBUG
    fprintf(stderr,"xmms_charset_init: ");
    if (lng) printf("lang: %s, ", lng);
    else printf("lang: <NULL>, ");
    if (id3) printf("id3: %s, ", id3);
    else printf("id3: <NULL>, ");
    if (output) printf("o: %s, ", output);
    else printf("o: <NULL>, ");
    if (fs) printf("fs: %s, ", fs);
    else printf("fs: <NULL>, ");
    if (pl) printf("pl: %s\n", pl);
    else printf("pl: <NULL>\n");
#endif
    default_language=get_default_language();
#ifdef DYNAMIC_ENCA 
    charset_dynamic_enca_init();
#endif
    if ((lng)&&(strcmp(lng,"Off")==0)) xmms_charset_new_language(0);
    else if ((lng)&&(strcmp(lng,"Default")==0)) xmms_charset_new_language(1);
    else { 
	if (lng) {
	    for (i=0;languages[i].sn;i++) {
		if (strcmp(languages[i].sn, lng)==0) break;
	    }
	}
        if ((i>=0)&&(languages[i].sn)) xmms_charset_new_language(i+2);
	else if (default_language<0) xmms_charset_new_language(0);
	else xmms_charset_new_language(default_language+2);
    }

    if (am<0) {
	if (autocharset_list_ni>1) am=1;
	else am=0;
    }
    xmms_autocharset_init(am,afm,afs);

    xmms_charset_id3 = xmms_charset_number(id3);
    xmms_charset_output = xmms_charset_number(output);
    xmms_charset_fs = xmms_charset_number(fs);
    xmms_charset_pl = xmms_charset_number(pl);

}

void xmms_charset_free() {
    xmms_autocharset_free();
    xmms_charset_iconv_close();
}


void xmms_charset_new_language(int scl) {
    int cl;

    if (scl==0) cl=-2;
    else if (scl==1) cl=default_language;
    else cl=scl-2;
    
    xmms_force_new = 0;
    if (cl==current_language) return;

#ifdef _CHARSET_DEBUG
    fprintf(stderr,"xmms_charset_new_language: %i\n",cl);
#endif

    current_language=cl;

#ifdef AUTO_ENCA_SUPPORT
    autocharset_enca_free();
#endif

    if ((scl==1)||(current_language==default_language)) xmms_charset_setdefault_charset();
    else strcpy(charset_default,"UTF-8");

    if (current_language>=0) {
	charset_list=languages[current_language].enc;
	autocharset_list_ni=languages[current_language].n_autoenc;
	autocharset_list=languages[current_language].autoenc;
    }
#ifdef AUTO_ENCA_SUPPORT
    autocharset_enca_init(current_language);
#endif
    xmms_charset_changed=1;
    xmms_autocharset_changed=1;
    xmms_force_new = 1;
}

void xmms_autocharset_new_mode(int m, int fs, int fm) {
    if ((!xmms_force_new)&&(xmms_autocharset_mode==m)&&(xmms_autocharset_fs==fs)&&(xmms_charset_fm==fm)) return;
#ifdef _CHARSET_DEBUG
    fprintf(stderr,"xmms_autocharset_new_mode: m: %d, fs: %d, fm: %d\n",m,fs,fm);
#endif

    xmms_autocharset_mode=m;
    xmms_autocharset_fs=fs;
    xmms_charset_fm=fm;
    xmms_autocharset_changed=1;
    xmms_charset_changed=1;
}

void xmms_charset_new_charsets(int id3, int output, int fs, int pl) {
    if ((!xmms_force_new)&&(xmms_charset_id3==id3)&&(xmms_charset_output==output)&&(xmms_charset_fs==fs)&&(xmms_charset_pl==pl)) return;
#ifdef _CHARSET_DEBUG
    fprintf(stderr,"xmms_charset_new_mode: %s %s %s %s\n", charset_list[id3], charset_list[output], charset_list[fs], charset_list[pl]);
#endif

    xmms_autocharset_changed = 1;
    xmms_charset_changed = 1;
    xmms_charset_id3 = id3;
    xmms_charset_output = output;
    xmms_charset_fs = fs;
    xmms_charset_pl = pl;
}

/*
PluginOptions default_po = {
    &xmms_charset_recode_fsout,
    NULL,
    NULL
};

void *charset_set_current_options(PluginOptions *cpo) {
    po=cpo;
}

PluginOptions *charset_get_current_options() {
    if (po) return po;
    else return &default_po;
}
*/


/*****************************************************************************
************************ Auto Recoding ***************************************
*****************************************************************************/
static int xmms_autocharset_number_charsets(int m) {
    charset_t *encodings;
    int i;

    if (!m) m=xmms_autocharset_mode;
    encodings=autocharset_list[m].encodings;
    for (i=0;encodings[i];i++);
    return i;
}


static void xmms_autocharset_iconv_close() {
    int i;
#ifdef _CHARSET_DEBUG
    fprintf(stderr,"xmms_autocharset_iconv_close\n");
#endif
    
    for (i=0;i<xmms_autocharset_number;i++) {
	if ((xmms_autocharset)&&(xmms_autocharset[i] != (iconv_t)-1)&&(xmms_autocharset[i] != (iconv_t)-2))
	    iconv_close(xmms_autocharset[i]);
	if ((xmms_autocharset_id2utf)&&(xmms_autocharset_id2utf[i] != (iconv_t)-1)&&(xmms_autocharset_id2utf[i] != (iconv_t)-2))
	    iconv_close(xmms_autocharset_id2utf[i]);
	if ((xmms_autocharset_id2id)&&(xmms_autocharset_id2id[i] != (iconv_t)-1)&&(xmms_autocharset_id2id[i] != (iconv_t)-2))
	    iconv_close(xmms_autocharset_id2id[i]);
	if ((xmms_autocharset_id2fs)&&(xmms_autocharset_id2fs[i] != (iconv_t)-1)&&(xmms_autocharset_id2fs[i] != (iconv_t)-2))
	    iconv_close(xmms_autocharset_id2fs[i]);
	if ((xmms_autocharset_fs2id)&&(xmms_autocharset_fs2id[i] != (iconv_t)-1)&&(xmms_autocharset_fs2id[i] != (iconv_t)-2))
	    iconv_close(xmms_autocharset_fs2id[i]);
    }
    if (xmms_autocharset) free(xmms_autocharset);
    if (xmms_autocharset_id2utf) free(xmms_autocharset_id2utf);
    if (xmms_autocharset_id2id) free(xmms_autocharset_id2id);
    if (xmms_autocharset_id2fs) free(xmms_autocharset_id2fs);
    if (xmms_autocharset_fs2id) free(xmms_autocharset_fs2id);
    
    xmms_autocharset=NULL;
    xmms_autocharset_id2utf=NULL;
    xmms_autocharset_id2id=NULL;
    xmms_autocharset_id2fs=NULL;
    xmms_autocharset_fs2id=NULL;
    xmms_autocharset_number = -1;
}





static void xmms_autocharset_iconv_open() {
    charset_t *encodings;
    const char *output, *id3, *fs;
    int i;

    assert(xmms_autocharset_changed>=0);
    
    if ((!xmms_autocharset_changed)&&(!xmms_charset_changed)) return;
    xmms_autocharset_iconv_close();
    xmms_autocharset_changed=0;

#ifdef _CHARSET_DEBUG
    fprintf(stderr,"xmms_autocharset_iconv_open\n");
#endif

    xmms_autocharset_number = xmms_autocharset_number_charsets(0);
    if (!xmms_autocharset_number) return;
    xmms_autocharset_changed=-1;
    
    xmms_autocharset = (iconv_t*)malloc((xmms_autocharset_number+1)*sizeof(iconv_t));
    xmms_autocharset_id2utf = (iconv_t*)malloc((xmms_autocharset_number+1)*sizeof(iconv_t));
    xmms_autocharset_id2id = (iconv_t*)malloc((xmms_autocharset_number+1)*sizeof(iconv_t));
    xmms_autocharset_id2fs = (iconv_t*)malloc((xmms_autocharset_number+1)*sizeof(iconv_t));
    xmms_autocharset_fs2id = (iconv_t*)malloc((xmms_autocharset_number+1)*sizeof(iconv_t));
    if ((!xmms_autocharset)||(!xmms_autocharset_id2id)||(!xmms_autocharset_id2fs)||(!xmms_autocharset_fs2id)||(!xmms_autocharset_id2utf)) {
	if (xmms_autocharset) {
	    free(xmms_autocharset);
	    xmms_autocharset=NULL;
	}
	if (xmms_autocharset_id2id) {
	    free(xmms_autocharset_id2id);
	    xmms_autocharset_id2id=NULL;
	}
	if (xmms_autocharset_id2utf) {
	    free(xmms_autocharset_id2utf);
	    xmms_autocharset_id2utf=NULL;
	}
	if (xmms_autocharset_id2fs) {
	    free(xmms_autocharset_id2fs);
	    xmms_autocharset_id2fs=NULL;
	}
	if (xmms_autocharset_fs2id) {
	    free(xmms_autocharset_fs2id);
	    xmms_autocharset_fs2id=NULL;
	}
	xmms_autocharset_number = -1;
	perror("iconv: allocate memory");
	return;
    }

    
    if (xmms_charset_output) output=charset_list[xmms_charset_output];
    else output=charset_default;
    if (xmms_charset_id3) id3=charset_list[xmms_charset_id3];
    else id3=charset_default;
    if (xmms_charset_fs) fs=charset_list[xmms_charset_fs];
    else fs=output;

    encodings = autocharset_list[xmms_autocharset_mode].encodings;
    
    for (i=0;encodings[i];i++) {
	if (strcmp(output,encodings[i])) {
#ifdef _CHARSET_DEBUG
	    fprintf(stderr,"id2out: %i %s %s\n",i,encodings[i],output);
#endif
	    xmms_autocharset[i] = iconv_open(output,encodings[i]);
	} else {
	    xmms_autocharset[i] = (iconv_t)-2;
	}
	
	if (strcmp(id3,encodings[i])) {
#ifdef _CHARSET_DEBUG
	    fprintf(stderr,"id2id: %i %s %s\n",i,encodings[i],id3);
#endif
	    xmms_autocharset_id2id[i] = iconv_open(id3,encodings[i]);
	} else {
	    xmms_autocharset_id2id[i] = (iconv_t)-2;
	}
	
	if (strcmp("UTF-8",encodings[i])) {
#ifdef _CHARSET_DEBUG
	    fprintf(stderr,"id2utf: %i %s\n",i,encodings[i]);
#endif
	    xmms_autocharset_id2utf[i] = iconv_open("UTF-8",encodings[i]);
	} else {
	    xmms_autocharset_id2utf[i] = (iconv_t)-2;
	}
	
	if (strcmp(fs,encodings[i])) {
#ifdef _CHARSET_DEBUG
	    fprintf(stderr,"idfs: %i %s %s\n",i,encodings[i],fs);
#endif
	    xmms_autocharset_id2fs[i] = iconv_open(fs,encodings[i]);
	    xmms_autocharset_fs2id[i] = iconv_open(encodings[i],fs);
	} else {
	    xmms_autocharset_id2fs[i] = (iconv_t)-2;
	    xmms_autocharset_fs2id[i] = (iconv_t)-2;
	    
	}
    }
    xmms_autocharset_changed=0;
}


static iconv_t xmms_autocharset_get(char *buf, int len) {
    int i;
    
    assert(buf);
    if (!buf) return (iconv_t)-1;
    
    xmms_autocharset_iconv_open();
    if (xmms_autocharset_changed<0) return (iconv_t)-1;
    if ((xmms_autocharset)&&(xmms_autocharset_number>0)) {
	i=autocharset_list[xmms_autocharset_mode].func(buf,len);
#ifdef _CHARSET_DEBUG
	fprintf(stderr,"xmms_autocharset_get: %i\n",i);
#endif
	if (i<xmms_autocharset_number) return xmms_autocharset[i];
    }
    return (iconv_t)-1;
}

static iconv_t xmms_autocharset_get_id2id(char *buf, int len) {
    int i;
    xmms_autocharset_iconv_open();
    if (xmms_autocharset_changed<0) return (iconv_t)-1;
    if ((xmms_autocharset_id2id)&&(xmms_autocharset_number>0)) {
	i=autocharset_list[xmms_autocharset_mode].func(buf,len);
#ifdef _CHARSET_DEBUG
	fprintf(stderr,"xmms_autocharset_get_id2id: %i\n",i);
#endif
	if (i<xmms_autocharset_number) return xmms_autocharset_id2id[i];
    }
    return (iconv_t)-1;
}

static iconv_t xmms_autocharset_get_id2utf(char *buf, int len) {
    int i;
    xmms_autocharset_iconv_open();
    if (xmms_autocharset_changed<0) return (iconv_t)-1;
    if ((xmms_autocharset_id2utf)&&(xmms_autocharset_number>0)) {
	i=autocharset_list[xmms_autocharset_mode].func(buf,len);
	if (i<xmms_autocharset_number) return xmms_autocharset_id2utf[i];
    }
    return (iconv_t)-1;
}

static iconv_t xmms_autocharset_get_id2fs(char *buf, int len) {
    int i;
    xmms_autocharset_iconv_open();
    if (xmms_autocharset_changed<0) return (iconv_t)-1;
    if ((xmms_autocharset_id2fs)&&(xmms_autocharset_number>0)) {
	i=autocharset_list[xmms_autocharset_mode].func(buf,len);
#ifdef _CHARSET_DEBUG
	fprintf(stderr,"xmms_autocharset_get_id2fs: %i\n",i);
#endif
	if (i<xmms_autocharset_number) return xmms_autocharset_id2fs[i];
    }
    return (iconv_t)-1;
}

#ifdef AUTO_ENCA_SUPPORT
static int autocharset_enca(char *buf, int len) {
    const char *charset;
    EncaEncoding ee;
    int i,l;

    if (!enca) return 0;
#ifdef _CHARSET_DEBUG
    fprintf(stderr,"xmms_autocharset_enca\n");
#endif
    l=strlen(buf);
    if ((len)&&(l>len)) l=len;
    ee=enca_analyse_const(enca,buf,l);
    if (ee.charset<0) return 0;
    charset=enca_charset_name(ee.charset, ENCA_NAME_STYLE_ICONV);
#ifdef _CHARSET_DEBUG
	fprintf(stderr,"xmms_autocharset_enca: %s\n",charset);
#endif
    for (i=0;autocharset_list[enca_pos].encodings[i];i++) {
	if (!strcmp(autocharset_list[enca_pos].encodings[i],charset)) return i;
    }
    return 0;
}
#endif

/*****************************************************************************
************************ Recoding ********************************************
*****************************************************************************/

static void xmms_charset_iconv_close() {
#ifdef _CHARSET_DEBUG
    fprintf(stderr,"xmms_charset_iconv_close\n");
#endif
    if (iconv_id3 != (iconv_t)-1) {
	if (iconv_close(iconv_id3)<0) perror("iconv close: id3");
	iconv_id3 = (iconv_t)-1;
    }
    if (iconv_output != (iconv_t)-1) {
	if (iconv_close(iconv_output)<0) perror("iconv close: output");
	iconv_output = (iconv_t)-1;
    }
    if (iconv_fs != (iconv_t)-1) {
	if (iconv_close(iconv_fs)<0) perror("iconv close: fs");
	iconv_fs = (iconv_t)-1;
    }
    if (iconv_fsout != (iconv_t)-1) {
	if (iconv_close(iconv_fsout)<0) perror("iconv close: fs2out");
	iconv_fsout = (iconv_t)-1;
    }
    if (iconv_fs2pl != (iconv_t)-1) {
	if (iconv_close(iconv_fs2pl)<0) perror("iconv close: fs2pl");
	iconv_fs2pl = (iconv_t)-1;
    }
    if (iconv_pl2fs != (iconv_t)-1) {
	if (iconv_close(iconv_pl2fs)<0) perror("iconv close: pl2fs");
	iconv_pl2fs = (iconv_t)-1;
    }
    if (iconv_utf != (iconv_t)-1) {
	if (iconv_close(iconv_utf)<0) perror("iconv close: utf");
	iconv_utf=(iconv_t)-1;
    }
    if (iconv_ct2out != (iconv_t)-1) {
	if (iconv_close(iconv_ct2out)<0) perror("iconv close: ct2out");
	iconv_ct2out=(iconv_t)-1;
    }
}

static void xmms_charset_iconv_open() {
    const char *id3, *output, *fs, *pl;
    assert(xmms_charset_changed>=0);
    if (!xmms_charset_changed) return;

#ifdef _CHARSET_DEBUG
    fprintf(stderr,"xmms_charset_iconv_open\n");
#endif

    xmms_charset_iconv_close();
    if (current_language<-1) return;

    if (xmms_charset_id3) id3=charset_list[xmms_charset_id3];
    else id3=charset_default;
    if (xmms_charset_output) output=charset_list[xmms_charset_output];
    else output=charset_default;
    if (xmms_charset_fs) fs=charset_list[xmms_charset_fs];
    else fs=output;
    if (xmms_charset_pl) pl=charset_list[xmms_charset_pl];
    else pl=id3;
    
    if (sure_default) {
	if (strcmp(charset_default,output)) {
	    iconv_ct2out = iconv_open(output,charset_default);
	}
    }
    
    
    if ((xmms_charset_id3 != xmms_charset_output)&&(strcmp(id3,output))) {
	iconv_id3 = iconv_open(output, id3);
	iconv_output = iconv_open(id3, output);
    }
    
    if (strcmp(fs,id3)) {
	iconv_fs = iconv_open(id3, fs);
    }
    if ((xmms_charset_fs)&&(xmms_charset_fs!=xmms_charset_output)) {
	iconv_fsout = iconv_open(output, fs);
    }

    if (strcmp(fs,pl)) {
	iconv_fs2pl = iconv_open(pl, fs);
	iconv_pl2fs = iconv_open(fs,pl);
    }
    
    if (strcmp(id3, "UTF-8")) {
	iconv_utf = iconv_open("UTF-8",id3);
    }

    if (strcmp(fs, "UTF-8")) {
	recode_fs2utf=1;
    } else {
	recode_fs2utf=0;
    }
    
    xmms_charset_changed = 0;
}


// Multi-byte strings still unsupported
static void iconv_copysymbol(char **in_buf, int *in_left, char **out_buf, int *out_left) {
    if ((out_left>0)&&(in_left>0)) {
	(**out_buf)=(**in_buf);
	(*out_buf)++;
	(*in_buf)++;
	(*in_left)--;
	(*out_left)--;
    }
}

static int utf_bytes(unsigned char c) {
    int j;
    if (c<128) return 1;

    for (j=6;j>=0;j--)
	if ((c&bit(j))==0) break;
	    
    if ((j==0)||(j==6)) return 1;
    return 6-j;
}

static char *xmms_charset_recode(iconv_t icnv, char *buf, int len, int *rlen) {
    char *in_buf, *out_buf, *res, err;
    int in_left, out_left, olen;
    int ub, utf_mode=0;
    int errors=0;
    
    assert(buf);
    if (!buf) return NULL;
    
    if (!len) {
	len=strlen(buf);
    } else {
	olen=strlen(buf);
	if (olen<len) len=olen;
    }

    if (iconv(icnv, NULL, NULL, NULL, NULL) == -1) {
	perror( "iconv convert: initialize" );
        return NULL;
    }

loop_restart:
    errors = 0;
    in_buf = buf;
    in_left = len;
    out_buf = xmms_charset_tmp;
    out_left = CHARSET_MAX_STRING_SIZE;

loop:
    err=iconv(icnv, &in_buf, &in_left, &out_buf, &out_left);
//    printf("%s, %i %u\n",buf,err,errno);
    if (err<0) {
        if (errno==E2BIG) {
    	    *(int*)(xmms_charset_tmp+(CHARSET_MAX_STRING_SIZE-sizeof(int)))=0;
	} else if (errno==EILSEQ) {
	    if (errors++<CHARSET_MAX_ERRORS) {
		if (utf_mode) {
		    for (ub=utf_bytes(*in_buf);ub>=0;ub--)
			iconv_copysymbol(&in_buf, &in_left, &out_buf, &out_left);
		}
		else {
//		    printf("%i: %s.\n",len-in_left,in_buf);
		    iconv_copysymbol(&in_buf, &in_left, &out_buf, &out_left);
		}
		if (in_left>0) goto loop;
	    } else if (!utf_mode) {
		utf_mode=1;
		goto loop_restart;
	    } else {
#ifdef _CHARSET_DEBUG
		fprintf(stderr,"Can't convert: %s.\n",buf);
#endif
	        perror("iconv convert: invalid encoding?");
	        return NULL;
	    }
	} else {
	    perror("iconv convert: convert");
	    return NULL;
	}
    }
        
    olen = CHARSET_MAX_STRING_SIZE - out_left;
    res = g_malloc(olen+1);
    if (!res) {
	perror( "iconv convert: malloc" );
	return NULL;
    }

    memcpy(res,xmms_charset_tmp,olen);
    res[olen]=0;
    if (rlen) *rlen=olen;
    return res;
}

char *xmms_charset_recode_ct2out(char *buf, int len, int *rlen) {
    if (current_language<-1) return NULL;

#ifdef _CHARSET_DEBUG
    fprintf(stderr,"xmms_charset_recode_ct2out\n");
#endif

    assert(buf);
    if (!buf) return NULL;
    
    xmms_charset_iconv_open();
    if (iconv_ct2out != (iconv_t)-1) 
	return xmms_charset_recode(iconv_ct2out, buf, len, rlen);
    return NULL;
}

char *xmms_charset_recode_id3(char *buf, int len, int *rlen) {
    iconv_t icnv;

    if (current_language<-1) return NULL;

#ifdef _CHARSET_DEBUG
    fprintf(stderr,"xmms_charset_recode_id3\n");
#endif
    
    assert(buf);
    if (!buf) return NULL;

    xmms_charset_iconv_open();
    xmms_autocharset_iconv_open();

    icnv = xmms_autocharset_get(buf,len);

    if (icnv == (iconv_t)-2) return NULL;

    if (icnv != (iconv_t)-1)
	return xmms_charset_recode(icnv, buf, len, rlen);

    if (iconv_id3 != (iconv_t)-1) 
	return xmms_charset_recode(iconv_id3, buf, len, rlen);
    return NULL;
}

char *xmms_charset_recode_output(char *buf, int len, int *rlen) {
    if (current_language<-1) return NULL;

#ifdef _CHARSET_DEBUG
    fprintf(stderr,"xmms_charset_recode_output\n");
#endif

    assert(buf);
    if (!buf) return NULL;
    
    xmms_charset_iconv_open();
    if (iconv_output != (iconv_t)-1) 
	return xmms_charset_recode(iconv_output, buf, len, rlen);
    return NULL;
}

char *xmms_charset_recode_fs(char *buf, int len, int *rlen) {
    iconv_t icnv;
    char *tmp1, *tmp2;

    if (current_language<-1) return NULL;
    
#ifdef _CHARSET_DEBUG
    fprintf(stderr,"xmms_charset_recode_fs\n");
#endif
    
    assert(buf);
    if (!buf) return NULL;

    xmms_charset_iconv_open();

    if (xmms_autocharset_fs) {
	icnv = xmms_autocharset_get(buf,len);
	if (icnv != (iconv_t)-1) {
	    if (icnv != (iconv_t)-2) {
		tmp1=xmms_charset_recode(icnv, buf, len, rlen);
		if (tmp1) {
		    if (iconv_output != (iconv_t)-1) {
			tmp2=xmms_charset_recode(iconv_output,tmp1,0,rlen);
			free(tmp1);
			return tmp2;
		    } else return tmp1;
		}
	    }
	    else if (iconv_output!=(iconv_t)-1) return xmms_charset_recode(iconv_output,buf,len,rlen);
	    else return NULL;
	}
    }
    if (iconv_fs != (iconv_t)-1) 
	return xmms_charset_recode(iconv_fs, buf, len, rlen);
    return NULL;
}

char *xmms_charset_recode_fsout(char *buf, int len, int *rlen) {
    iconv_t icnv;

    if (current_language<-1) return NULL;

#ifdef _CHARSET_DEBUG
    fprintf(stderr,"xmms_charset_recode_fsout\n");
#endif

    assert(buf);
    if (!buf) return NULL;

    xmms_charset_iconv_open();

    if (xmms_autocharset_fs) {
	icnv = xmms_autocharset_get(buf,len);
	if (icnv == (iconv_t)-2) return NULL;
	if (icnv != (iconv_t)-1) {
	    return xmms_charset_recode(icnv, buf, len, rlen);
	}
    }
    if (iconv_fsout != (iconv_t)-1) 
	return xmms_charset_recode(iconv_fsout, buf, len, rlen);
    return NULL;
}

char *xmms_charset_recode_fs2pl(char *buf, int len, int *rlen) {
    iconv_t icnv;
    char *fs;
/*
    int i;
*/
    if (current_language<-1) return NULL;
    if (!xmms_charset_fm) return NULL;

#ifdef _CHARSET_DEBUG
    fprintf(stderr,"xmms_charset_recode_fs2pl\n");
#endif

    assert(buf);
    if (!buf) return NULL;

    xmms_charset_iconv_open();
/*
    if (xmms_autocharset_fs) {
	xmms_autocharset_iconv_open();
	if (xmms_charset_fs) fs=charset_list[xmms_charset_fs];
	else {
	    if (xmms_charset_output) fs=charset_list[xmms_charset_output];
	    else fs=charset_default;
	}
	i=autocharset_list[xmms_autocharset_mode].func(buf,len);
	if (i<xmms_autocharset_number) {
	    if (strcmp(charset_list[i],fs)) return NULL;
	}
    }
*/
    if (xmms_autocharset_fs) {
	icnv = xmms_autocharset_get_id2fs(buf,len);
	if (icnv!=(iconv_t)-2) {
	    if (icnv!=(iconv_t)-1) {
		fs=xmms_charset_recode(icnv,buf,len,rlen);
		if (fs) {
		    if (iconv_fs2pl != (iconv_t)-1) {
			buf=xmms_charset_recode(iconv_fs2pl, fs, 0, rlen);
			free(fs);
			return buf;
		    } else 
			return fs;
		} else return NULL;
	    }
	}
    }
        
    if (iconv_fs2pl != (iconv_t)-1) 
	return xmms_charset_recode(iconv_fs2pl, buf, len, rlen);
    return NULL;
}

char *xmms_charset_recode_pl2fs(char *buf, int len, int *rlen) {
    iconv_t icnv;

    if (current_language<-1) return NULL;
    if (!xmms_charset_fm) return NULL;

#ifdef _CHARSET_DEBUG
    fprintf(stderr,"xmms_charset_recode_pl2fs\n");
#endif

    assert(buf);
    if (!buf) return NULL;

    xmms_charset_iconv_open();
    
    icnv = xmms_autocharset_get_id2fs(buf,len);
	
    if (icnv==(iconv_t)-2) return NULL;

    if (icnv != (iconv_t)-1) {
	return xmms_charset_recode(icnv, buf, len, rlen);
    }
    if (iconv_pl2fs != (iconv_t)-1) {
	return xmms_charset_recode(iconv_pl2fs, buf, len, rlen);
    }
    return NULL;
}

static int check_file(char *path, char *name) {
    struct stat st;
    char *temp;
    
    temp = g_strdup_printf("%s/%s", path, name);
    if (!temp) return 0;
    stat(temp,&st);
    free(temp);
    
    if (S_ISREG(st.st_mode)) return 1;
    return 0;
}

char *xmms_charset_recode_fs2fs(char *path, char *buf) {
    int i,len;
    charset_t *encodings;
    char *tmp;

    if (current_language<-1) return NULL;
    if (!xmms_charset_fm) return NULL;
    if (!xmms_autocharset_fs) return NULL;

#ifdef _CHARSET_DEBUG
    fprintf(stderr,"xmms_charset_recode_fs2fs\n");
#endif
    
    assert(path);
    assert(buf);
    if ((!path)||(!buf)) return NULL;
    
    xmms_charset_iconv_open();
    xmms_autocharset_iconv_open();
    if (xmms_autocharset_changed<0) return NULL;
    if (!xmms_autocharset_fs2id) return NULL;

    if (check_file(path,buf)) return NULL;
    encodings = autocharset_list[xmms_autocharset_mode].encodings;
    
    len=strlen(buf);
    for (i=0;encodings[i];i++) {
	if ((xmms_autocharset_fs2id[i]!=(iconv_t)-1)&&(xmms_autocharset_fs2id[i]!=(iconv_t)-2)) {
	    tmp=xmms_charset_recode(xmms_autocharset_fs2id[i],buf,len,0);
	    if (tmp) {
		if (check_file(path,tmp)) return tmp;
		free(tmp);
	    }
	}
    }
    return NULL;
}

char *xmms_charset_recode_id2utf(char *buf, int len, int *rlen) {
    char *str = NULL;
    int olen;
    iconv_t icnv;

    if (current_language<-1) return NULL;

    xmms_charset_iconv_open();

#ifdef _CHARSET_DEBUG
    fprintf(stderr,"xmms_charset_recode_id2utf\n");
#endif
    
    assert(buf);
    if (!buf) return NULL;

    icnv = xmms_autocharset_get_id2id(buf,len);
    if ((icnv != (iconv_t)-1)&&(icnv != (iconv_t)-2)) {
	str = xmms_charset_recode(icnv, buf, len, &olen);
	if (!str) return NULL;
	len = olen;
    }
    
    xmms_charset_iconv_open();
    if (iconv_utf != (iconv_t)-1) {
	buf=xmms_charset_recode(iconv_utf, str ? str : buf, len, rlen);
	if (str) free(str);
	return buf;
    }

    return NULL;
}

char *xmms_charset_recode_fs2utf(char *buf, int len, int *rlen) {
    int olen;
    char *str = NULL;
    iconv_t icnv;

    if (current_language<-1) return NULL;

#ifdef _CHARSET_DEBUG
    fprintf(stderr,"xmms_charset_recode_fs2utf\n");
#endif

    assert(buf);
    if (!buf) return NULL;

    xmms_charset_iconv_open();

    if (xmms_autocharset_fs) {
	icnv = xmms_autocharset_get_id2utf(buf,len);
	if (icnv==(iconv_t)-2) return NULL;
	if (icnv!=(iconv_t)-1) {
	    str = xmms_charset_recode(icnv, buf, len, rlen);
	    if (str) return str;
	}
    }
        
    if (recode_fs2utf) {
	if (iconv_fs != (iconv_t)-1) {
	    str = xmms_charset_recode(iconv_fs, buf, len, &olen);
	    if (!str) return NULL;
	    len=olen;
	}
	if (iconv_utf != (iconv_t)-1) {
	    buf = xmms_charset_recode(iconv_utf, str ? str : buf, len, &olen);
	    if (str) free(str);
	} else if (str) {
	    buf = str;
	} else {
	    buf = NULL;
	}
	if (rlen) *rlen=olen;
	return buf;
    }
    return NULL;
}

