#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <glib.h>
#include <iconv.h>
#include <locale.h>
#include <assert.h>

/*
#include "../config.h"
#ifdef HAVE_CODESET
#include <langinfo.h>
#endif
*/

#include "charset.h"
#include "charset_config.h"

#define LOCALE_STRING_LENGTH 64
static char charset_default[LOCALE_STRING_LENGTH];
static int current_language;
static char xmms_charset_tmp[CHARSET_MAX_STRING_SIZE];

static int xmms_charset_changed = -1;

static int xmms_charset_id3;
static int xmms_charset_output;
static int xmms_charset_fs;
static int xmms_charset_pl;

static iconv_t iconv_id3=(iconv_t)-1;	/* id3 to output */
static iconv_t iconv_output=(iconv_t)-1; /* output to id3 */
static iconv_t iconv_fs=(iconv_t)-1; /* fs to id3 (depricated) */
static iconv_t iconv_fsout=(iconv_t)-1; /* fs to out */
static iconv_t iconv_utf=(iconv_t)-1; /* id3 to UTF-8 */
static iconv_t iconv_pl2fs=(iconv_t)-1; /* playlist to fs */
static iconv_t iconv_fs2pl=(iconv_t)-1; /* fs to playlist */
static iconv_t iconv_ct2out=(iconv_t)-1; /* LC_CTYPE to output */

static int xmms_autocharset_mode = 0;
static int xmms_autocharset_fs = 1;
static int xmms_autocharset_changed = 0;
static iconv_t *xmms_autocharset = NULL;
static iconv_t *xmms_autocharset_id2id = NULL;
static iconv_t *xmms_autocharset_id2fs = NULL;
static int xmms_autocharset_number = -1;

//static int recode_fsout;
static int recode_fs2utf;
#include "charset_auto_russian.h"

extern int errno;

static int sure_default;
static PluginOptions *po = NULL;

const char *charset_get_output_charset() {
    if (xmms_charset_output) return charset_list[xmms_charset_output];
    else return charset_default;
}

PluginOptions default_po = {
    &xmms_charset_recode_fsout,
    NULL,
    NULL
};

void *charset_set_current_options(PluginOptions *cpo) {
    po=cpo;
}

PluginOptions *charset_get_current_options() {
    if (po) return po;
    else return &default_po;
}

/*****************************************************************************
************************ Auto Recoding ***************************************
*****************************************************************************/
static int xmms_autocharset_number_charsets(int m) {
    const char **encodings;
    int i;

    if (!m) m=xmms_autocharset_mode;
    encodings=autocharset_list[m].encodings;
    for (i=0;encodings[i];i++);
    return i;
}


static void xmms_autocharset_iconv_close() {
    int i;
    
    for (i=0;i<xmms_autocharset_number;i++) {
	if ((xmms_autocharset)&&(xmms_autocharset[i] != (iconv_t)-1)&&(xmms_autocharset[i] != (iconv_t)-2))
	    iconv_close(xmms_autocharset[i]);
	if ((xmms_autocharset_id2id)&&(xmms_autocharset_id2id[i] != (iconv_t)-1)&&(xmms_autocharset_id2id[i] != (iconv_t)-2))
	    iconv_close(xmms_autocharset_id2id[i]);
	if ((xmms_autocharset_id2fs)&&(xmms_autocharset_id2fs[i] != (iconv_t)-1)&&(xmms_autocharset_id2fs[i] != (iconv_t)-2))
	    iconv_close(xmms_autocharset_id2fs[i]);
    }
    if (xmms_autocharset) free(xmms_autocharset);
    if (xmms_autocharset_id2id) free(xmms_autocharset_id2id);
    if (xmms_autocharset_id2fs) free(xmms_autocharset_id2fs);
    
    xmms_autocharset=NULL;
    xmms_autocharset_id2id=NULL;
    xmms_autocharset_id2fs=NULL;
    xmms_autocharset_number = -1;
}

void xmms_autocharset_new_mode(int m, int fs) {
    xmms_autocharset_mode=m;
    xmms_autocharset_fs=fs;
    xmms_autocharset_changed=1;
}

void xmms_autocharset_init(int m) {
    xmms_autocharset_mode=m;
    xmms_autocharset_changed=1;
}

void xmms_autocharset_free() {
    xmms_autocharset_iconv_close();
}

static void xmms_autocharset_iconv_open() {
    const char **encodings;
    const char *output, *id3, *fs;
    int i;
    
    if ((!xmms_autocharset_changed)&&(!xmms_charset_changed)) return;
    xmms_autocharset_iconv_close();
    xmms_autocharset_changed=0;

    xmms_autocharset_number = xmms_autocharset_number_charsets(0);
    if (!xmms_autocharset_number) return;
    
    xmms_autocharset = (iconv_t*)malloc((xmms_autocharset_number+1)*sizeof(iconv_t));
    xmms_autocharset_id2id = (iconv_t*)malloc((xmms_autocharset_number+1)*sizeof(iconv_t));
    xmms_autocharset_id2fs = (iconv_t*)malloc((xmms_autocharset_number+1)*sizeof(iconv_t));
    if ((!xmms_autocharset)||(!xmms_autocharset_id2id)||(!xmms_autocharset_id2fs)) {
	if (xmms_autocharset) {
	    free(xmms_autocharset);
	    xmms_autocharset=NULL;
	}
	if (xmms_autocharset_id2id) {
	    free(xmms_autocharset_id2id);
	    xmms_autocharset_id2id=NULL;
	}
	if (xmms_autocharset_id2fs) {
	    free(xmms_autocharset_id2fs);
	    xmms_autocharset_id2fs=NULL;
	}
	xmms_autocharset_number = -1;
	perror("iconv: allocate memory");
	return;
    }

    
    if (xmms_charset_output) output=charset_list[xmms_charset_output];
    else output=charset_default;
    if (xmms_charset_id3) id3=charset_list[xmms_charset_id3];
    else id3=charset_default;
    if (xmms_charset_fs) fs=charset_list[xmms_charset_fs];
    else fs=output;

    encodings = autocharset_list[xmms_autocharset_mode].encodings;
    
    for (i=0;encodings[i];i++) {
	if (strcmp(output,encodings[i]))
	    xmms_autocharset[i] = iconv_open(output,encodings[i]);
	else
	    xmms_autocharset[i] = (iconv_t)-2;
	
	if (strcmp(id3,encodings[i]))
	    xmms_autocharset_id2id[i] = iconv_open(id3,encodings[i]);
	else 
	    xmms_autocharset_id2id[i] = (iconv_t)-2;
	
	if (strcmp(fs,encodings[i])) 
	    xmms_autocharset_id2fs[i] = iconv_open(fs,encodings[i]);
	else
	    xmms_autocharset_id2fs[i] = (iconv_t)-2;
    }
}


static iconv_t xmms_autocharset_get(char *buf, int len) {
    int i;
    xmms_autocharset_iconv_open();
    if ((xmms_autocharset)&&(xmms_autocharset_number>0)) {
	i=autocharset_list[xmms_autocharset_mode].func(buf,len);
	if (i<xmms_autocharset_number) return xmms_autocharset[i];
    }
    return (iconv_t)-1;
}

static iconv_t xmms_autocharset_get_id2id(char *buf, int len) {
    int i;
    xmms_autocharset_iconv_open();
    if ((xmms_autocharset_id2id)&&(xmms_autocharset_number>0)) {
	i=autocharset_list[xmms_autocharset_mode].func(buf,len);
	if (i<xmms_autocharset_number) return xmms_autocharset_id2id[i];
    }
    return (iconv_t)-1;
}

static iconv_t xmms_autocharset_get_id2fs(char *buf, int len) {
    int i;
    xmms_autocharset_iconv_open();
    if ((xmms_autocharset_id2fs)&&(xmms_autocharset_number>0)) {
	i=autocharset_list[xmms_autocharset_mode].func(buf,len);
	if (i<xmms_autocharset_number) return xmms_autocharset_id2fs[i];
    }
    return (iconv_t)-1;
}

/*****************************************************************************
************************ Recoding ********************************************
*****************************************************************************/

static void xmms_charset_iconv_close() {
    if (iconv_id3 != (iconv_t)-1) {
	if (iconv_close(iconv_id3)<0) perror("iconv close: id3");
	iconv_id3 = (iconv_t)-1;
    }
    if (iconv_output != (iconv_t)-1) {
	if (iconv_close(iconv_output)<0) perror("iconv close: output");
	iconv_output = (iconv_t)-1;
    }
    if (iconv_fs != (iconv_t)-1) {
	if (iconv_close(iconv_fs)<0) perror("iconv close: fs");
	iconv_fs = (iconv_t)-1;
    }
    if (iconv_fsout != (iconv_t)-1) {
	if (iconv_close(iconv_fsout)<0) perror("iconv close: fs2out");
	iconv_fsout = (iconv_t)-1;
    }
    if (iconv_fs2pl != (iconv_t)-1) {
	if (iconv_close(iconv_fs2pl)<0) perror("iconv close: fs2pl");
	iconv_fs2pl = (iconv_t)-1;
    }
    if (iconv_pl2fs != (iconv_t)-1) {
	if (iconv_close(iconv_pl2fs)<0) perror("iconv close: pl2fs");
	iconv_pl2fs = (iconv_t)-1;
    }
    if (iconv_utf != (iconv_t)-1) {
	if (iconv_close(iconv_utf)<0) perror("iconv close: utf");
	iconv_utf=(iconv_t)-1;
    }
    if (iconv_ct2out != (iconv_t)-1) {
	if (iconv_close(iconv_ct2out)<0) perror("iconv close: ct2out");
	iconv_ct2out=(iconv_t)-1;
    }
}

static void xmms_charset_iconv_open() {
    const char *id3, *output, *fs, *pl;
    assert(xmms_charset_changed>=0);
    if (!xmms_charset_changed) return;

    xmms_charset_iconv_close();
    if (xmms_charset_id3) id3=charset_list[xmms_charset_id3];
    else id3=charset_default;
    if (xmms_charset_output) output=charset_list[xmms_charset_output];
    else output=charset_default;
    if (xmms_charset_fs) fs=charset_list[xmms_charset_fs];
    else fs=output;
    if (xmms_charset_pl) pl=charset_list[xmms_charset_pl];
    else pl=id3;
    
    if (sure_default) {
	if (strcmp(charset_default,output)) {
	    iconv_ct2out = iconv_open(output,charset_default);
	}
    }
    
    
    if ((xmms_charset_id3 != xmms_charset_output)&&(strcmp(id3,output))) {
	iconv_id3 = iconv_open(output, id3);
	iconv_output = iconv_open(id3, output);
    }
    
    if (strcmp(fs,id3)) {
	iconv_fs = iconv_open(id3, fs);
    }
    if ((xmms_charset_fs)&&(xmms_charset_fs!=xmms_charset_output)) {
	iconv_fsout = iconv_open(output, fs);
    }

    if (strcmp(fs,pl)) {
	iconv_fs2pl = iconv_open(pl, fs);
	iconv_pl2fs = iconv_open(fs,pl);
    }
    
    if (strcmp(id3, "UTF-8")) {
	iconv_utf = iconv_open("UTF-8",id3);
    }

    if (strcmp(fs, "UTF-8")) {
	recode_fs2utf=1;
    } else {
	recode_fs2utf=0;
    }
    
    xmms_charset_changed = 0;
}


static int xmms_charset_number(char *c) {
    int i;
    if (!c) return 0;
    for (i=0;charset_list[i];i++)
	if (strcmp(charset_list[i], c)==0) return i;
    return 0; /* Default, - if not found */
}

static void xmms_charset_setdefault() {
    char *l;
    char ltmp[3];
    int i;

	// Detecting default charset    
/*
My routine is better!
#ifdef HAVE_CODESET
    char *cs;
    cs=nl_langinfo(CODESET);
    if (cs) {
	strncpy(charset_default,cs,LOCALE_STRING_LENGTH);
	charset_default[LOCALE_STRING_LENGTH-1]=0;
    } else {
	strcpy(charset_default,"UTF-8");
    }
#else
*/
    char *str1, *str2;
    str1 = setlocale(ENV_VARIABLE,"");
    if (str1) {
        str2 = strrchr(str1,'.');
        if (str2) strncpy(charset_default,str2+1,LOCALE_STRING_LENGTH);
	else strncpy(charset_default,str1,LOCALE_STRING_LENGTH);
	charset_default[LOCALE_STRING_LENGTH-1]=0;
        str2 = strchr(charset_default,'@');
        if (str2) *str2 = 0;
	sure_default=1;
    } else {
        strcpy(charset_default,"UTF-8");
	sure_default=0;
    }
/* #endif */

    ltmp[2]=0;

	// Detecting language
    l=setlocale(ENV_VARIABLE,"");
    if (!l) current_language=-1;
    else if ((strcmp(l,"C")==0)||(strcmp(l,"POSIX")==0)) current_language=-1;
    else {
	for (i=0;language_aliases[i].alias;i++) 
	    if (strncmp(l,language_aliases[i].alias,5)==0) {
		strncpy(ltmp,language_aliases[i].lang,2);
		break;
	    }
	if (!language_aliases[i].alias) strncpy(ltmp,l,2);
	for (i=0;languages[i];i++)
	    if (strcmp(ltmp,languages[i])==0) {
		current_language=i;
		break;
	    }
	if (!languages[i]) current_language=-1;
    }
    
    if (current_language>=0) charset_list=charset_language+current_language;
}

void xmms_charset_init(char *id3, char *output, char *fs, char *pl) {
    if (xmms_charset_changed<0) xmms_charset_setdefault();
	
    xmms_charset_changed = 1;
    xmms_charset_id3 = xmms_charset_number(id3);
    xmms_charset_output = xmms_charset_number(output);
    xmms_charset_fs = xmms_charset_number(fs);
    xmms_charset_pl = xmms_charset_number(pl);
}

void xmms_charset_free() {
    xmms_charset_iconv_close();
}

void xmms_charset_new_charsets(int id3, int output, int fs, int pl) {
    if (xmms_charset_changed<0) 
	xmms_charset_setdefault();
    else 
	if ((xmms_charset_id3==id3)&&(xmms_charset_output==output)&&(xmms_charset_fs==fs)&&(xmms_charset_pl==pl)) return;

    xmms_charset_changed = 1;
    xmms_charset_id3 = id3;
    xmms_charset_output = output;
    xmms_charset_fs = fs;
    xmms_charset_pl = pl;
}

// Multi-byte strings still unsupported
static void iconv_copysymbol(char **in_buf, int *in_left, char **out_buf, int *out_left) {
    if ((out_left>0)&&(in_left>0)) {
	(**out_buf)=(**in_buf);
	(*out_buf)++;
	(*in_buf)++;
	(*in_left)--;
	(*out_left)--;
    }
}

static char *xmms_charset_recode(iconv_t icnv, char *buf, int len, int *rlen) {
    char *in_buf, *out_buf, *res, err;
    int in_left, out_left, olen;
    int errors=0;
    
    assert(buf);
    
    if (!len) {
	len=strlen(buf);
    } else {
	olen=strlen(buf);
	if (olen<len) len=olen;
    }

    if (iconv(icnv, NULL, NULL, NULL, NULL) == -1) {
	perror( "iconv convert: initialize" );
        return NULL;
    }

    in_buf = buf;
    in_left = len;
    out_buf = &xmms_charset_tmp;
    out_left = CHARSET_MAX_STRING_SIZE;

loop:
    err=iconv(icnv, &in_buf, &in_left, &out_buf, &out_left);
//    printf("%s, %i %u\n",buf,err,errno);
    if (err<0) {
        if (errno==E2BIG) {
    	    *(int*)(xmms_charset_tmp+(CHARSET_MAX_STRING_SIZE-sizeof(int)))=0;
	} else if (errno==EILSEQ) {
	    if (errors<CHARSET_MAX_ERRORS) {
	        errors++;
		iconv_copysymbol(&in_buf, &in_left, &out_buf, &out_left);
		if (in_left>0) goto loop;
	    } else {
	        perror("iconv convert: invalid encoding?");
	        return NULL;
	    }
	} else {
	    perror("iconv convert: convert");
//	    perror(buf);
	    return NULL;
	}
    }
        
    olen = CHARSET_MAX_STRING_SIZE - out_left;
    res = g_malloc(olen+1);
    if (!res) {
	perror( "iconv convert: malloc" );
	return NULL;
    }

    memcpy(res,xmms_charset_tmp,olen);
    res[olen]=0;
    if (rlen) *rlen=olen;
    return res;
}

char *xmms_charset_recode_ct2out(char *buf, int len, int *rlen) {
    xmms_charset_iconv_open();
    if (iconv_ct2out != (iconv_t)-1) 
	return xmms_charset_recode(iconv_ct2out, buf, len, rlen);
    return NULL;
}

char *xmms_charset_recode_id3(char *buf, int len, int *rlen) {
    iconv_t icnv;

    xmms_charset_iconv_open();
    icnv = xmms_autocharset_get(buf,len);

    if (icnv == (iconv_t)-2) return NULL;

    if (icnv != (iconv_t)-1)
	return xmms_charset_recode(icnv, buf, len, rlen);

    if (iconv_id3 != (iconv_t)-1) 
	return xmms_charset_recode(iconv_id3, buf, len, rlen);
    return NULL;
}

char *xmms_charset_recode_output(char *buf, int len, int *rlen) {
    xmms_charset_iconv_open();
    if (iconv_output != (iconv_t)-1) 
	return xmms_charset_recode(iconv_output, buf, len, rlen);
    return NULL;
}

char *xmms_charset_recode_fs(char *buf, int len, int *rlen) {
    xmms_charset_iconv_open();
    if (iconv_fs != (iconv_t)-1) 
	return xmms_charset_recode(iconv_fs, buf, len, rlen);
    return NULL;
}

char *xmms_charset_recode_fsout(char *buf, int len, int *rlen) {
    xmms_charset_iconv_open();
    if (iconv_fsout != (iconv_t)-1) 
	return xmms_charset_recode(iconv_fsout, buf, len, rlen);
    return NULL;
}

char *xmms_charset_recode_fs2pl(char *buf, int len, int *rlen) {
    xmms_charset_iconv_open();
    if (iconv_fs2pl != (iconv_t)-1) 
	return xmms_charset_recode(iconv_fs2pl, buf, len, rlen);
    return NULL;
}

char *xmms_charset_recode_pl2fs(char *buf, int len, int *rlen) {
    iconv_t icnv;

    xmms_charset_iconv_open();
    icnv = xmms_autocharset_get_id2fs(buf,len);
	
    if (icnv==(iconv_t)-2) return NULL;

    if (icnv != (iconv_t)-1)
	return xmms_charset_recode(icnv, buf, len, rlen);

    if (iconv_pl2fs != (iconv_t)-1) 
	return xmms_charset_recode(iconv_pl2fs, buf, len, rlen);
    return NULL;
}


char *xmms_charset_recode_id2utf(char *buf, int len, int *rlen) {
    char *str = NULL;
    int olen;
    iconv_t icnv;

    xmms_charset_iconv_open();

    icnv = xmms_autocharset_get_id2id(buf,len);
    if ((icnv != (iconv_t)-1)&&(icnv != (iconv_t)-2)) {
	str = xmms_charset_recode(icnv, buf, len, &olen);
	if (!str) return NULL;
	len = olen;
    }
    
    xmms_charset_iconv_open();
    if (iconv_utf != (iconv_t)-1) {
	buf=xmms_charset_recode(iconv_utf, str ? str : buf, len, rlen);
	if (str) free(str);
	return buf;
    }

    return NULL;
}

char *xmms_charset_recode_fs2utf(char *buf, int len, int *rlen) {
    int olen;
    char *str = NULL;

    xmms_charset_iconv_open();
    if (recode_fs2utf) {
	if (iconv_fs != (iconv_t)-1) {
	    str = xmms_charset_recode(iconv_fs, buf, len, &olen);
	    if (!str) return NULL;
	    len=olen;
	}
	if (iconv_utf != (iconv_t)-1) {
	    buf = xmms_charset_recode(iconv_utf, str ? str : buf, len, &olen);
	    if (str) free(str);
	} else if (str) {
	    buf = str;
	} else {
	    buf = NULL;
	}
	if (rlen) *rlen=olen;
	return buf;
    }
    return NULL;
}
