#ifndef XMMS_CHARSET_H
#define XMMS_CHARSET_H

#include "charset_defines.h"

int autocharset_russian(char *buf,int len);
int autocharset_russian_uc(char *buf,int len);

#ifdef AUTO_ENCA_SUPPORT
int autocharset_enca_russian(char *buf, int len);
#endif

typedef char *(*recode_function)(char *buf, int len, int *rlen);

struct plugin_options_t {
    recode_function fs;
    recode_function tag;
    recode_function after;    
};

typedef struct plugin_options_t PluginOptions;

struct language_alias_t {
    char *alias;
    char *lang;
};

struct autocharset_list_t {
    char *title;
    int (*func)(char *buf, int len);
    char *encodings[CHARSET_MAX_ENCODINGS];
};

typedef const char *charset_list_t[CHARSET_MAX_ENCODINGS];

extern const char **charset_list;
extern const int autocharset_list_ni;
extern const struct autocharset_list_t autocharset_list[];

/* Data Support functions */
const char *charset_get_output_charset();
void *charset_set_current_options(PluginOptions *cpo);
PluginOptions *charset_get_current_options();

/* Initialization functions */
void xmms_autocharset_new_mode(int m, int fs);
void xmms_autocharset_init(int m);
void xmms_autocharset_free();

void xmms_charset_init(char *id3, char *output, char *fs, char *pl);
void xmms_charset_free();
void xmms_charset_new_charsets(int id3, int output, int fs, int pl);

/* Recoding functions */
char *xmms_charset_recode_ct2out(char *buf, int len, int *rlen);
char *xmms_charset_recode_id3(char *buf, int len, int *rlen);
char *xmms_charset_recode_output(char *buf, int len, int *rlen);
char *xmms_charset_recode_fs(char *buf, int len, int *rlen);
char *xmms_charset_recode_fsout(char *buf, int len, int *rlen);
char *xmms_charset_recode_id2utf(char *buf, int len, int *rlen);
char *xmms_charset_recode_fs2utf(char *buf, int len, int *rlen);
char *xmms_charset_recode_fs2pl(char *buf, int len, int *rlen);
char *xmms_charset_recode_pl2fs(char *buf, int len, int *rlen);

#endif /* XMMS_CHARSET_H */
