#include <stdio.h>
#include <locale.h>
#include <string.h>
#include <gtk/gtk.h>
#include "xmms.h"
#include "libxmms/charset.h"
#include "charset.h"

#define CHARSET_MAX_STRING_SIZE	1024	/* 256 symbols in UTF-32 */
static char xmms_charset_tmp[CHARSET_MAX_STRING_SIZE];

extern TextBox *playlistwin_sinfo;

static gint selected_id3_charset, selected_output_charset, selected_fs_charset;
static gint selected_pl_charset;
static gint selected_autocharset;
static GtkWidget *prefswin_options_font_shade_entry, *prefswin_options_font_shade_browse;

static void prefswin_font_shade_browse_ok(GtkWidget * w, gpointer data)
{
	GtkFontSelectionDialog *fontsel = GTK_FONT_SELECTION_DIALOG(data);
	gchar *fontname;

	fontname = gtk_font_selection_dialog_get_font_name(fontsel);
	
	if (fontname)
		gtk_entry_set_text(GTK_ENTRY(prefswin_options_font_shade_entry), fontname);

	gtk_widget_destroy(GTK_WIDGET(fontsel));
}

static void prefswin_font_shade_browse_cb(GtkWidget * w, gpointer data)
{
	static GtkWidget *fontsel;

	if (fontsel != NULL)
		return;

	fontsel = gtk_font_selection_dialog_new(_("Select playlist font:"));
	gtk_font_selection_dialog_set_font_name(GTK_FONT_SELECTION_DIALOG(fontsel), gtk_entry_get_text(GTK_ENTRY(prefswin_options_font_shade_entry)));
	gtk_signal_connect(GTK_OBJECT(GTK_FONT_SELECTION_DIALOG(fontsel)->ok_button), "clicked", GTK_SIGNAL_FUNC(prefswin_font_shade_browse_ok), fontsel);
	gtk_signal_connect_object(GTK_OBJECT(GTK_FONT_SELECTION_DIALOG(fontsel)->cancel_button), "clicked", GTK_SIGNAL_FUNC(gtk_widget_destroy), GTK_OBJECT(fontsel));
	gtk_signal_connect(GTK_OBJECT(fontsel), "destroy", GTK_SIGNAL_FUNC(gtk_widget_destroyed), &fontsel);
	gtk_widget_show(fontsel);
}


static void prefswin_charset_cb(GtkWidget * w, gpointer item)
{
    int i;
    i = GPOINTER_TO_INT(item);
    if (i>=0x3FFFF) selected_pl_charset=i-0x3FFFF;
    else if (i>=0x2FFFF) selected_autocharset=i-0x2FFFF;
    else if (i>=0x1FFFF) selected_fs_charset=i-0x1FFFF;
    else if (i>=0xFFFF) selected_output_charset=i-0xFFFF;
    else selected_id3_charset=i;
}

static void charset_options_add(Config *cfg, GtkOptionMenu *cmenu1, GtkOptionMenu *cmenu2, GtkOptionMenu *cmenu3, GtkOptionMenu *cmenu4)
{
    int i;
    GtkWidget *menu1, *item1;
    GtkWidget *menu2, *item2;
    GtkWidget *menu3, *item3;
    GtkWidget *menu4, *item4;

    menu1 = gtk_menu_new();
    gtk_option_menu_remove_menu(cmenu1);
    gtk_option_menu_set_menu(cmenu1, menu1);

    menu2 = gtk_menu_new();
    gtk_option_menu_remove_menu(cmenu2);
    gtk_option_menu_set_menu(cmenu2, menu2);

    menu3 = gtk_menu_new();
    gtk_option_menu_remove_menu(cmenu3);
    gtk_option_menu_set_menu(cmenu3, menu3);

    menu4 = gtk_menu_new();
    gtk_option_menu_remove_menu(cmenu4);
    gtk_option_menu_set_menu(cmenu4, menu4);

    selected_id3_charset=0;
    selected_output_charset=0;
    selected_fs_charset=0;
    selected_pl_charset=0;
    
    for (i=0;charset_list[i];i++) {
	item1 = gtk_menu_item_new_with_label(charset_list[i]);
	gtk_signal_connect(GTK_OBJECT(item1), "activate", GTK_SIGNAL_FUNC(prefswin_charset_cb), GINT_TO_POINTER(i));
	gtk_widget_show(item1);
        gtk_menu_append(GTK_MENU(menu1), item1);
	item2 = gtk_menu_item_new_with_label(charset_list[i]);
	gtk_signal_connect(GTK_OBJECT(item2), "activate", GTK_SIGNAL_FUNC(prefswin_charset_cb), GINT_TO_POINTER(i+0xFFFF));
	gtk_widget_show(item2);
        gtk_menu_append(GTK_MENU(menu2), item2);
	item3 = gtk_menu_item_new_with_label(charset_list[i]);
	gtk_signal_connect(GTK_OBJECT(item3), "activate", GTK_SIGNAL_FUNC(prefswin_charset_cb), GINT_TO_POINTER(i+0x1FFFF));
	gtk_widget_show(item3);
        gtk_menu_append(GTK_MENU(menu3), item3);
	item4 = gtk_menu_item_new_with_label(charset_list[i]);
	gtk_signal_connect(GTK_OBJECT(item4), "activate", GTK_SIGNAL_FUNC(prefswin_charset_cb), GINT_TO_POINTER(i+0x3FFFF));
	gtk_widget_show(item4);
        gtk_menu_append(GTK_MENU(menu4), item4);
	if ((cfg->charset_id3)&&(!strcmp(cfg->charset_id3,charset_list[i]))) {
	    selected_id3_charset=i;
	}
	if ((cfg->charset_output)&&(!strcmp(cfg->charset_output,charset_list[i]))) {
	    selected_output_charset=i;
	}
	if ((cfg->charset_fs)&&(!strcmp(cfg->charset_fs,charset_list[i]))) {
	    selected_fs_charset=i;
	}
	if ((cfg->charset_pl)&&(!strcmp(cfg->charset_pl,charset_list[i]))) {
	    selected_pl_charset=i;
	}
    }

    gtk_option_menu_set_history(cmenu1, selected_id3_charset);
    gtk_option_menu_set_history(cmenu2, selected_output_charset);
    gtk_option_menu_set_history(cmenu3, selected_fs_charset);
    gtk_option_menu_set_history(cmenu4, selected_pl_charset);
}

static void autocharset_options_add(Config *cfg, GtkOptionMenu *cmenu)
{
    int i;
    GtkWidget *menu, *item;

    menu = gtk_menu_new();
    gtk_option_menu_remove_menu(cmenu);
    gtk_option_menu_set_menu(cmenu, menu);

    selected_autocharset=0;
    
    
    for (i=0;i<autocharset_list_ni;i++) {
	item = gtk_menu_item_new_with_label(autocharset_list[i].title);
	gtk_signal_connect(GTK_OBJECT(item), "activate", GTK_SIGNAL_FUNC(prefswin_charset_cb), GINT_TO_POINTER(i+0x2FFFF));
	gtk_widget_show(item);
        gtk_menu_append(GTK_MENU(menu), item);
	if ((cfg->autocharset)&&(cfg->autocharset==i)) {
	    selected_autocharset=i;
	}
    }

    gtk_option_menu_set_history(cmenu, selected_autocharset);
}

static GtkWidget *prefswin_autocharset_fs_option;
void charset_create_prefs_window(Config *cfg, GtkNotebook *prefswin_notebook, GtkWidget *prefswin_fonts_vbox) {
	GtkWidget *prefswin_encodings_vbox;
	GtkWidget *prefswin_fonts_shade_frame;
	GtkWidget *options_font_shade_hbox, *options_font_shade_vbox;
	GtkWidget *prefswin_charset_frame, *prefswin_charset_vbox;
	GtkWidget *prefswin_charset_list1, *prefswin_charset_list2, *prefswin_charset_list3, *prefswin_charset_list4;
	GtkWidget *prefswin_charset_hbox1, *prefswin_charset_hbox2, *prefswin_charset_hbox3, *prefswin_charset_hbox4;
	GtkWidget *prefswin_charset_label1, *prefswin_charset_label2, *prefswin_charset_label3, *prefswin_charset_label4;
	GtkWidget *prefswin_autocharset_frame, *prefswin_autocharset_vbox;
	GtkWidget *prefswin_autocharset_list;
	GtkWidget *prefswin_autocharset_hbox;
	GtkWidget *prefswin_autocharset_label;
 	GtkWidget *prefswin_autocharset_fs_hbox;

	prefswin_fonts_shade_frame = gtk_frame_new(_("Shade"));
	gtk_container_set_border_width(GTK_CONTAINER(prefswin_fonts_shade_frame), 5);
	gtk_box_pack_start(GTK_BOX(prefswin_fonts_vbox), prefswin_fonts_shade_frame, FALSE, FALSE, 0);
	options_font_shade_vbox = gtk_vbox_new(FALSE, 5);
	gtk_container_border_width(GTK_CONTAINER(options_font_shade_vbox), 5);
	gtk_container_add(GTK_CONTAINER(prefswin_fonts_shade_frame), options_font_shade_vbox);
	options_font_shade_hbox = gtk_hbox_new(FALSE, 5);
	gtk_box_pack_start_defaults(GTK_BOX(options_font_shade_vbox), options_font_shade_hbox);
	prefswin_options_font_shade_entry = gtk_entry_new();
	gtk_box_pack_start(GTK_BOX(options_font_shade_hbox), prefswin_options_font_shade_entry, TRUE, TRUE, 0);
	prefswin_options_font_shade_browse = gtk_button_new_with_label(_("Browse"));
	gtk_signal_connect(GTK_OBJECT(prefswin_options_font_shade_browse), "clicked", GTK_SIGNAL_FUNC(prefswin_font_shade_browse_cb), NULL);
	gtk_widget_set_usize(prefswin_options_font_shade_browse, 85, 17);
	gtk_box_pack_start(GTK_BOX(options_font_shade_hbox), prefswin_options_font_shade_browse, FALSE, TRUE, 0);

	prefswin_encodings_vbox = gtk_vbox_new(FALSE, 0);
	
	prefswin_charset_frame = gtk_frame_new(_("Encodings"));
	gtk_box_pack_start(GTK_BOX(prefswin_encodings_vbox), prefswin_charset_frame, FALSE, FALSE, 0);
	gtk_container_border_width(GTK_CONTAINER(prefswin_charset_frame), 5);
	prefswin_charset_vbox = gtk_vbox_new(FALSE, 3);
	gtk_container_add(GTK_CONTAINER(prefswin_charset_frame), prefswin_charset_vbox);
	gtk_container_border_width(GTK_CONTAINER(prefswin_charset_vbox), 5);

	prefswin_charset_hbox1 = gtk_hbox_new(FALSE, 10);
	gtk_container_add(GTK_CONTAINER(prefswin_charset_vbox), prefswin_charset_hbox1);
	gtk_container_border_width(GTK_CONTAINER(prefswin_charset_hbox1), 5);
	prefswin_charset_label1 = gtk_label_new(_("ID3 Encoding"));
	gtk_widget_set_usize(prefswin_charset_label1, 120, 17);
	gtk_box_pack_start(GTK_BOX(prefswin_charset_hbox1), prefswin_charset_label1, FALSE, FALSE, 0);
	prefswin_charset_list1 = gtk_option_menu_new();
	gtk_box_pack_start(GTK_BOX(prefswin_charset_hbox1), prefswin_charset_list1, TRUE, TRUE, 0);

	prefswin_charset_hbox3 = gtk_hbox_new(FALSE, 10);
	gtk_container_add(GTK_CONTAINER(prefswin_charset_vbox), prefswin_charset_hbox3);
	gtk_container_border_width(GTK_CONTAINER(prefswin_charset_hbox3), 5);
	prefswin_charset_label3 = gtk_label_new(_("FileSystem Encoding"));
	gtk_widget_set_usize(prefswin_charset_label3, 120, 17);
	gtk_box_pack_start(GTK_BOX(prefswin_charset_hbox3), prefswin_charset_label3, FALSE, FALSE, 0);
	prefswin_charset_list3 = gtk_option_menu_new();
	gtk_box_pack_start(GTK_BOX(prefswin_charset_hbox3), prefswin_charset_list3, TRUE, TRUE, 0);

	prefswin_charset_hbox4 = gtk_hbox_new(FALSE, 10);
	gtk_container_add(GTK_CONTAINER(prefswin_charset_vbox), prefswin_charset_hbox4);
	gtk_container_border_width(GTK_CONTAINER(prefswin_charset_hbox4), 5);
	prefswin_charset_label4 = gtk_label_new(_("Playlist Encoding"));
	gtk_widget_set_usize(prefswin_charset_label4, 120, 17);
	gtk_box_pack_start(GTK_BOX(prefswin_charset_hbox4), prefswin_charset_label4, FALSE, FALSE, 0);
	prefswin_charset_list4 = gtk_option_menu_new();
	gtk_box_pack_start(GTK_BOX(prefswin_charset_hbox4), prefswin_charset_list4, TRUE, TRUE, 0);

	prefswin_charset_hbox2 = gtk_hbox_new(FALSE, 10);
	gtk_container_add(GTK_CONTAINER(prefswin_charset_vbox), prefswin_charset_hbox2);
	gtk_container_border_width(GTK_CONTAINER(prefswin_charset_hbox2), 5);
	prefswin_charset_label2 = gtk_label_new(_("Output Encoding"));
	gtk_widget_set_usize(prefswin_charset_label2, 120, 17);
	gtk_box_pack_start(GTK_BOX(prefswin_charset_hbox2), prefswin_charset_label2, FALSE, FALSE, 0);
	prefswin_charset_list2 = gtk_option_menu_new();
	gtk_box_pack_start(GTK_BOX(prefswin_charset_hbox2), prefswin_charset_list2, TRUE, TRUE, 0);

	charset_options_add(cfg, GTK_OPTION_MENU(prefswin_charset_list1),GTK_OPTION_MENU(prefswin_charset_list2),GTK_OPTION_MENU(prefswin_charset_list3), GTK_OPTION_MENU(prefswin_charset_list4));

	prefswin_autocharset_frame = gtk_frame_new(_("Encoding Autoselection"));
	gtk_box_pack_start(GTK_BOX(prefswin_encodings_vbox), prefswin_autocharset_frame, FALSE, FALSE, 0);
	gtk_container_border_width(GTK_CONTAINER(prefswin_autocharset_frame), 5);
	prefswin_autocharset_vbox = gtk_vbox_new(FALSE, 3);
	gtk_container_add(GTK_CONTAINER(prefswin_autocharset_frame), prefswin_autocharset_vbox);
	gtk_container_border_width(GTK_CONTAINER(prefswin_autocharset_vbox), 5);

	prefswin_autocharset_hbox = gtk_hbox_new(FALSE, 10);
	gtk_container_add(GTK_CONTAINER(prefswin_autocharset_vbox), prefswin_autocharset_hbox);
	gtk_container_border_width(GTK_CONTAINER(prefswin_autocharset_hbox), 5);
	prefswin_autocharset_label = gtk_label_new(_("Language"));
	gtk_widget_set_usize(prefswin_autocharset_label, 120, 17);
	gtk_box_pack_start(GTK_BOX(prefswin_autocharset_hbox), prefswin_autocharset_label, FALSE, FALSE, 0);
	prefswin_autocharset_list = gtk_option_menu_new();
	gtk_box_pack_start(GTK_BOX(prefswin_autocharset_hbox), prefswin_autocharset_list, TRUE, TRUE, 0);

	autocharset_options_add(cfg, GTK_OPTION_MENU(prefswin_autocharset_list));


 	prefswin_autocharset_fs_hbox = gtk_hbox_new(FALSE, 10);
 	gtk_container_add(GTK_CONTAINER(prefswin_autocharset_vbox), prefswin_autocharset_fs_hbox);
 	gtk_container_border_width(GTK_CONTAINER(prefswin_autocharset_fs_hbox), 5);
 	prefswin_autocharset_fs_option = gtk_check_button_new_with_label(_("Enable for FileSystem"));	
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(prefswin_autocharset_fs_option),cfg->autocharset_fs);
 	gtk_box_pack_start_defaults(GTK_BOX(prefswin_autocharset_fs_hbox), prefswin_autocharset_fs_option);

	gtk_notebook_append_page(prefswin_notebook, prefswin_encodings_vbox, gtk_label_new(_("Encodings")));
}

void charset_prefswin_apply_changes(Config *cfg) {
	if (cfg->charset_id3) g_free(cfg->charset_id3);
	if (cfg->charset_output) g_free(cfg->charset_output);
	if (cfg->charset_fs) g_free(cfg->charset_fs);
	if (cfg->charset_pl) g_free(cfg->charset_pl);
	cfg->charset_id3=g_strdup(charset_list[selected_id3_charset]);
	cfg->charset_output=g_strdup(charset_list[selected_output_charset]);
	cfg->charset_fs=g_strdup(charset_list[selected_fs_charset]);
	cfg->charset_pl=g_strdup(charset_list[selected_pl_charset]);
	xmms_charset_new_charsets(selected_id3_charset, selected_output_charset, selected_fs_charset, selected_pl_charset);
	cfg->autocharset=selected_autocharset;
	cfg->autocharset_fs=gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(prefswin_autocharset_fs_option));
	xmms_autocharset_new_mode(selected_autocharset,cfg->autocharset_fs);

	if (cfg->shade_font) g_free(cfg->shade_font);
	cfg->shade_font = g_strdup(gtk_entry_get_text(GTK_ENTRY(prefswin_options_font_shade_entry)));
        textbox_set_xfont(playlistwin_sinfo, cfg->mainwin_use_xfont, cfg->shade_font);
}

void charset_show_prefs_window(Config *cfg) {
	gtk_entry_set_text(GTK_ENTRY(prefswin_options_font_shade_entry), cfg->shade_font);
}

void charset_create_playlist_window(Config *cfg) {
	textbox_set_xfont(playlistwin_sinfo, cfg->mainwin_use_xfont, cfg->shade_font);
}
/*****************************************************************************
************************ Initialization / Deinitilization ********************
*****************************************************************************/
void charset_free(Config *cfg) {
    xmms_charset_free();
    xmms_autocharset_free();
}

void charset_init(Config *cfg) {
    xmms_charset_init(cfg->charset_id3, cfg->charset_output, cfg->charset_fs, cfg->charset_pl);
    xmms_autocharset_init(cfg->autocharset);
}

void charset_read_config(ConfigFile *cfgfile, Config *cfg) {
	cfg->autocharset=0;
	cfg->autocharset_fs=1;
	xmms_cfg_read_string(cfgfile, "xmms", "shade_font", &cfg->shade_font);
	xmms_cfg_read_string(cfgfile, "xmms", "charset_id3", &cfg->charset_id3);
	xmms_cfg_read_string(cfgfile, "xmms", "charset_output", &cfg->charset_output);
	xmms_cfg_read_string(cfgfile, "xmms", "charset_fs", &cfg->charset_fs);
	xmms_cfg_read_string(cfgfile, "xmms", "charset_pl", &cfg->charset_pl);
	xmms_cfg_read_int(cfgfile,"xmms","autocharset",&cfg->autocharset);
	xmms_cfg_read_int(cfgfile,"xmms","autocharset_fs",&cfg->autocharset_fs);
 	if (cfg->autocharset>autocharset_list_ni) cfg->autocharset=0;
}

void charset_write_config(ConfigFile *cfgfile, Config *cfg) {
	xmms_cfg_write_string(cfgfile, "xmms", "shade_font", cfg->shade_font);
	xmms_cfg_write_string(cfgfile, "xmms", "charset_id3", cfg->charset_id3);
	xmms_cfg_write_string(cfgfile, "xmms", "charset_output", cfg->charset_output);
	xmms_cfg_write_string(cfgfile, "xmms", "charset_fs", cfg->charset_fs);
	xmms_cfg_write_string(cfgfile, "xmms", "charset_pl", cfg->charset_pl);
	xmms_cfg_write_int(cfgfile,"xmms","autocharset",cfg->autocharset);
	xmms_cfg_write_int(cfgfile,"xmms","autocharset_fs",cfg->autocharset_fs);
}

void charset_default_config(Config *cfg) {
	if (cfg->shade_font && strlen(cfg->shade_font) == 0)
	{
		g_free(cfg->shade_font);
		cfg->shade_font = NULL;
	}
	if (cfg->shade_font == NULL)
		cfg->shade_font = g_strdup("-misc-fixed-medium-r-*-*-7-*");
}

/*****************************************************************************
*********************************** Sort *************************************
*****************************************************************************/
gchar *saved_locale;

static void charset_sort_null_entry(gpointer entry, gpointer data) {
    ((PlaylistEntry*)entry)->sort=NULL;
}

static void charset_sort_free_entry(gpointer entry, gpointer data) {
    if (((PlaylistEntry*)entry)->sort)
	g_free(((PlaylistEntry*)entry)->sort);
}

void charset_sort_prepare(GList *playlist) {
    g_list_foreach(playlist, charset_sort_null_entry, NULL);
    saved_locale=g_strdup(setlocale(LC_COLLATE,NULL));
    setlocale(LC_COLLATE,"UTF-8");
}

void charset_sort_free(GList *playlist) {
    g_list_foreach(playlist, charset_sort_free_entry, NULL);
    if (saved_locale) {
	setlocale(LC_COLLATE,saved_locale);
	g_free(saved_locale);
    }
}

char *charset_sort_convert(PlaylistEntry *entry, gchar *title, int fs) {
    char *str;
    if (entry->sort) return entry->sort;
    if (fs)
	str=xmms_charset_recode_fs2utf(title,0,NULL);
    else
	str=xmms_charset_recode_id2utf(title,0,NULL);

    if (!str)
	entry->sort=g_strdup(title);
    else {
	strxfrm(xmms_charset_tmp,str,CHARSET_MAX_STRING_SIZE);
//    	*(int*)(xmms_charset_tmp+(CHARSET_MAX_STRING_SIZE-sizeof(int)))=0;
	xmms_charset_tmp[CHARSET_MAX_STRING_SIZE-1]=0;
	entry->sort=g_strdup(xmms_charset_tmp);
	g_free(str);
    }

    if (entry->sort)
	return entry->sort;
	
    return title;
}

/*****************************************************************************
*********************************** Misc *************************************
*****************************************************************************/

void textbox_set_ctext(TextBox * tb, gchar * text)
{
	gchar *rtext;
	
        lock_widget(tb);
	
	rtext=xmms_charset_recode_id3(text,0,NULL);
	if (rtext) text=rtext;
	if (tb->tb_text)
	{
		if (!strcmp(text, tb->tb_text))
		{
			if (rtext) free(rtext);
		        unlock_widget(tb);
			return;
		}
		g_free(tb->tb_text);
	}

	tb->tb_text = g_strdup(text);
	if (rtext) free(rtext);

	unlock_widget(tb);
	draw_widget(tb);
}
